{
  Original Name: Twitter_StringComposer.pas
  Copyright: Copyright (c) 2010, Simon J Stuart, All Rights Reserved
  License: Licensed under the MIT License.
  Author: Simon J Stuart
  Author's Website: http://www.lakraven.com
  Description:
    This unit contains Classes/Methods etc. necessary to produce
    our API URLs, Parameters, and any other dynamically-constructed
    String data. This is just to make the flow of code in TTwitter
    cleaner and easier to follow.

  --------------------------------------------------------------------------------
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
  --------------------------------------------------------------------------------
  CODE NOTES:
}
unit Twitter_StringComposer;

interface

uses
  Classes, SysUtils,
  Twitter_Common;

type
  TTwitterParam = record
    Key: string;
    Value: string;
    IsFile: Boolean;
  public
    function ToString: string;
  end;

  {
    TTwitterParamFactory object is basically an Array Manager for POST Header Parameters.
  }
  TTwitterParamFactory = class(TObject)
  private
    FIsMultiPart: Boolean;
    FParams: TArray<TTwitterParam>;
    FCount: Integer;
    function GetParam(const AIndex: Integer): TTwitterParam;
  public
    constructor Create;
    procedure Add(const AKey, AValue: string; AIsFile: Boolean = False);
    function ToString: string; override;
    property Count: Integer read FCount;
    property IsMultiPart: Boolean read FIsMultiPart;
    property Params[const AIndex: Integer]: TTwitterParam read GetParam; default;
  end;

  {
    TTwitterURLAssembler object provides methods to produce our API URLs (and POST Parameters, where applicable)
    This exists merely to unify the way URL strings are generated, also avoids repeated
    calls to the "Format()" function in our API-invoking routines.
    Just to make it clear, this entire class exists as "Syntactic-Sugar"
  }
  TTwitterURLAssembler = class(TObject)
  public
    function OAuth(const OAuthParam: TTwitterOAuthParams): String;
    function UsersShow(const Username: String; const AIncludeEntities: Boolean): String; overload; // TO RETRIEVE A USER'S DETAILS FROM THEIR NAME
    function UsersShow(const UserID: Int64; const AIncludeEntities: Boolean): String; overload; // TO RETRIEVE A USER'S DETAILS FROM THEIR ID <- MORE EFFICIENT
    function Timeline(const ATimeline: TTwitterTimelineType; const ACount: Integer = 0; const ASince_id: Int64 = 0; const AMax_id: Int64 = 0; const APage: Integer = 0; const AScreenName: String = ''; const AIncludeEntities: Boolean = True; const ATrimUser: Boolean = False; const ARetweets: Boolean = True; const AUser_id: Int64 = 0): String;
    function PostTweet(const AMsg: String; const AReplyToStatusID: Int64; const AImage: string; const ALat, ALong: Extended; const APlaceID: String; const ADisplayCoordinates: Boolean; ATrimUser, AIncludeEntities: Boolean; var AParams: TTwitterParamFactory): String;
    function RateLimitStatus: String;
  end;

  {
    TTwitterParameterPair object simply holds a Key and Value
  }
  TTwitterParameterPair = record
    Key: String;
    Value: String;
  end;
  TTwitterParameterPairArray = Array of TTwitterParameterPair;

  {
    TTwitterStringSplitter object provides methods to take URL Parameter Strings and split them into Key:Value pairs!
    This object exists to make the core implementation neater and easier to read.
  }
  TTwitterStringSplitter = class(TObject)
  private
    FCount: Integer;
    FParams: TTwitterParameterPairArray;
    function Explode(const ADelimiter: String; const AValue: String; const ACount: Integer) : TTwitterStringArray;
    function GetIndex(const AName: String): Integer;
    function GetValue(const AName: String): String;
  public
    function OrderURL(const AURL: String): String;
    procedure ParseURLParams(const AURLParams: String);
    property Count: Integer read FCount;
    property Value[const AName: String]: String read GetValue; default;
  end;

implementation

{ TTwitterParam }

function TTwitterParam.ToString: string;
begin
  Result := Format('%s=%s', [Key, Value]);
end;

{ TTwitterParamFactory }

procedure TTwitterParamFactory.Add(const AKey, AValue: string; AIsFile: Boolean = False);
var
  LIndex: Integer;
begin
  LIndex := Length(FParams);
  SetLength(FParams, LIndex + 1);
  with FParams[LIndex] do
  begin
    Key := AKey;
    Value := AValue;
    IsFile := AIsFile;
  end;
  FIsMultiPart := FIsMultiPart or AIsFile;
  Inc(FCount);
end;

constructor TTwitterParamFactory.Create;
begin
  inherited;
  FCount := 0;
end;

function TTwitterParamFactory.GetParam(const AIndex: Integer): TTwitterParam;
begin
  Result := FParams[AIndex];
end;

function TTwitterParamFactory.ToString: string;
var
  i: Integer;
begin
  Result := '';
  if Length(FParams) > 0 then
    Result := Format('%s=%s', [FParams[0].Key, FParams[0].Value]);
  for i := Low(FParams)+1 to High(FParams) do
    Result := Result + Format('&%s=%s', [FParams[i].Key, FParams[i].Value]);
end;

{ TTwitterURLAssembler }

function TTwitterURLAssembler.OAuth(const OAuthParam: TTwitterOAuthParams): String;
const
  URL = 'https://api.twitter.com/oauth/%s';
  TWITTER_OAUTH_STRINGS: Array[TTwitterOAuthParams] of String = ('request_token', 'access_token', 'authorize', 'authenticate');
begin
  Result := Format(URL, [TWITTER_OAUTH_STRINGS[OAuthParam]]);
end;

function TTwitterURLAssembler.PostTweet(const AMsg: String; const AReplyToStatusID: Int64; const AImage: string; const ALat, ALong: Extended; const APlaceID: String; const ADisplayCoordinates: Boolean; ATrimUser, AIncludeEntities: Boolean; var AParams: TTwitterParamFactory): String;
const
  TWITTER_URL_STATUSES_UPDATE = 'statuses/update.json';
begin
  if AParams = nil then
    AParams := TTwitterParamFactory.Create;

  AParams.Add('status', AMsg);

  if AReplyToStatusID > 0 then
    AParams.Add('in_reply_to_status_id', IntToStr(AReplyToStatusID));
  if (ALat >= -90) and (ALat <= 90) then
    AParams.Add('lat', FloatToStr(ALat));
  if (ALong >= -180) and (ALong <= 180) then
    AParams.Add('long', FloatToStr(ALong));
  if APlaceID <> EmptyStr then
    AParams.Add('place_id', APlaceID);
  if ADisplayCoordinates then
    AParams.Add('display_coordinates', 'true');
  if ATrimUser then
    AParams.Add('trim_user', 'true');
  if AIncludeEntities then
    AParams.Add('include_entities', 'true');

  if (AImage <> '') and FileExists(AImage) then
  begin
    AParams.Add('media[]', AImage, True);
    Result := 'https://upload.twitter.com/1/statuses/update_with_media.json'
  end
  else
    Result := Format(TWITTER_URL_BASE, [TWITTER_API_VERSION, TWITTER_URL_STATUSES_UPDATE]);
end;

function TTwitterURLAssembler.RateLimitStatus: String;
const
  TWITTER_URL_ACCOUNT_RATELIMIT = 'account/rate_limit_status.json';
begin
  Result := Format(TWITTER_URL_BASE, [TWITTER_API_VERSION, TWITTER_URL_ACCOUNT_RATELIMIT]);
end;

function TTwitterURLAssembler.Timeline(const ATimeline: TTwitterTimelineType;
  const ACount: Integer = 0; const ASince_id: Int64 = 0; const AMax_id: Int64 = 0;
  const APage: Integer = 0; const AScreenName: String = '';
  const AIncludeEntities: Boolean = True; const ATrimUser: Boolean = False;
  const ARetweets: Boolean = True; const AUser_id: Int64 = 0): String;
const
  TWITTER_URL_TIMELINES: Array [TTwitterTimelineType] of String = (
                                                                 '',
                                                                 'statuses/friends_timeline.json%s',
                                                                 'statuses/home_timeline.json%s',
                                                                 'statuses/mentions.json%s',
                                                                 'statuses/public_timeline.json%s',
                                                                 'statuses/retweeted_by_me.json%s',
                                                                 'statuses/retweeted_to_me.json%s',
                                                                 'statuses/retweets_of_me.json%s',
                                                                 'statuses/user_timeline.json%s'
                                                               );
var
  Timeline,
  LParams: String;
begin
  Timeline := TWITTER_URL_TIMELINES[ATimeline];
  LParams := EmptyStr;
  if ASince_id > 0 then
    LParams := LParams + Format('&since_id=%d', [ASince_id]);
  if ACount > 0 then
    LParams := LParams + Format('&max_id=%d', [AMax_id]);
  if APage > 0 then
    LParams := LParams + Format('&count=%d', [ACount]);
  if AMax_id > 0 then
    LParams := LParams + Format('&page=%d', [APage]);
  if AScreenName <> EmptyStr then
    LParams := LParams + Format('&screen_name=%s', [AScreenName]);
  if AUser_id > 0 then
    LParams := LParams + Format('&user_id=%d', [AUser_id]);

  if ATrimUser then
    LParams := LParams + '&trim_user=1';
  if AIncludeEntities then
    LParams := LParams + '&include_entities=true';
  if ARetweets then
    LParams := LParams + '&include_rts=true';
  if Length(LParams) > 1 then
    LParams[1] := '?';
  Result := Format(TWITTER_URL_BASE, [TWITTER_API_VERSION, Format(Timeline, [LParams])]);
end;

function TTwitterURLAssembler.UsersShow(const Username: String; const AIncludeEntities: Boolean): String;
const
  TWITTER_URL_USERS = 'users/show.json?screen_name=%s';
var
  LParams: String;
begin
  if AIncludeEntities then
    LParams := LParams + '&include_entities=true'
  else
    LParams := EmptyStr;
  Result := Format(TWITTER_URL_BASE, [TWITTER_API_VERSION, Format(TWITTER_URL_USERS, [Username])]) + LParams;
end;

function TTwitterURLAssembler.UsersShow(const UserID: Int64; const AIncludeEntities: Boolean): String;
const
  TWITTER_URL_USERS = 'users/show.json?user_id%d';
var
  LParams: String;
begin
  if AIncludeEntities then
    LParams := LParams + '&include_entities=true'
  else
    LParams := EmptyStr;
  Result := Format(TWITTER_URL_BASE, [TWITTER_API_VERSION, Format(TWITTER_URL_USERS, [UserID])]) + LParams;
end;

{ TTwitterStringSplitter }

function TTwitterStringSplitter.Explode(const ADelimiter, AValue: String; const ACount: Integer): TTwitterStringArray;
var
  s : String;
  i,
  p : Integer;
begin
  s := AValue; i := 0;
  while length(s) > 0 do
  begin
    inc(i);
    SetLength(result, i);
    p := pos(ADelimiter,s);
    if ( p > 0 ) and ( ( i < ACount ) OR ( ACount = 0) ) then
    begin
      result[i - 1] := copy(s,0,p-1);
      s := copy(s,p + length(ADelimiter),length(s));
    end else
    begin
      result[i - 1] := s;
      s := '';
    end;
  end;
end;

function TTwitterStringSplitter.GetIndex(const AName: String): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := Result + 1 to FCount - 1 do
  begin
    if FParams[I].Key = AName then
    begin
      Result := I;
      Break;
    end;
  end;
end;

function TTwitterStringSplitter.GetValue(const AName: String): String;
var
  LIdx: Integer;
begin
  LIdx := GetIndex(AName); // Determine the index of the named Parameter
  if LIdx = -1 then
    Result := EmptyStr  // If the index is -1, then no such parameter exists, just return an Empty String
  else
  begin
    Result := FParams[LIdx].Value; // Return the Value
  end;
end;

function TTwitterStringSplitter.OrderURL(const AURL: String): String;
var
  I: Integer;
  SL: TStringList;
  URL,
  Parameters: String;
begin
  Result := AURL;
  URL := Copy(AURL, 0, Pos('?', AURL) -1);
  Parameters := Copy(AURL, Length(URL) + 2, Length(AURL));
  if Parameters = EmptyStr then
    Exit;
  ParseURLParams(Copy(AURL, Pos('?', AURL) + 1, Length(AURL)));
  SL := TStringList.Create;
  try
    for I := 0 to FCount - 1 do
      SL.Add(FParams[I].Key + '=' + Self.FParams[I].Value);
    SL.Sort;
    for I := 0 to SL.Count - 1 do
    begin
      if I = 0 then
        Result := URL + '?' + SL[I]
      else
        Result := Result + '&' + SL[I];
    end;
  finally
    SL.Free;
  end;

end;

procedure TTwitterStringSplitter.ParseURLParams(const AURLParams: String);
var
  LParams,
  LPairs: TTwitterStringArray;
  I: Integer;
begin
  SetLength(FParams, 0);
  LParams := Explode('&', AURLParams, 0); // Populate the Parameter Array
  FCount := Length(LParams); // Update the Param Count
  for I := 0 to FCount - 1 do
  begin
    LPairs := Explode('=', LParams[I], 0);
    SetLength(FParams, Length(FParams) + 1);
    with FParams[Length(FParams) - 1] do
    begin
      Key := LPairs[0];
      if Length(LPairs) = 2 then
        Value := LPairs[1]
      else
        Value := EmptyStr;
    end;
  end;

end;

end.
